package de.fraunhofer.iem.dataprovider.kpi.service

import de.fraunhofer.iem.dataprovider.kpi.dto.KPICreateDto
import de.fraunhofer.iem.dataprovider.kpi.entity.KPIEntity
import de.fraunhofer.iem.dataprovider.kpi.entity.KPIHierarchyEdgeEntity
import de.fraunhofer.iem.dataprovider.kpi.repository.KPIHierarchyEdgeRepository
import de.fraunhofer.iem.dataprovider.kpi.repository.KPIRepository
import de.fraunhofer.iem.dataprovider.logger.getLogger
import de.fraunhofer.iem.dataprovider.repository.entity.RepositoryEntity
import org.springframework.data.repository.findByIdOrNull
import org.springframework.stereotype.Service
import java.util.*

@Service
class KPIService(
    private val kpiRepository: KPIRepository,
    private val kpiHierarchyEdgeRepository: KPIHierarchyEdgeRepository
) {

    private val logger = getLogger(javaClass)

    private fun calculateKPIsRecursively(kpi: KPICreateDto, visited: MutableSet<KPICreateDto>) {
        // Check if the KPI has already been visited
        if (visited.contains(kpi)) {
            return
        }

        // Check if the KPI has child KPIs
        if (kpi.hierarchyEdges.isEmpty()) {
            // Leaf node, calculate the KPI value
            kpi.calculateKPI()
            visited.add(kpi)
            return
        }

        // Recursively calculate child KPIs first
        for (childEdge in kpi.hierarchyEdges) {
            calculateKPIsRecursively(childEdge.to, visited)
        }

        // Calculate the KPI value after processing child KPIs
        kpi.calculateKPI()
        visited.add(kpi)
    }

    fun calculateKPIs(rootKPI: KPICreateDto) {
        val visited: MutableSet<KPICreateDto> = mutableSetOf()
        calculateKPIsRecursively(rootKPI, visited)
    }

    fun storeAndPurgeOld(repositoryEntity: RepositoryEntity, rootKPI: KPICreateDto) {
        purgeAllKPIs(repositoryEntity)
        val kpiEntityDtoToEntityMapping: MutableMap<KPICreateDto, KPIEntity> = mutableMapOf()
        storeKPI(repositoryEntity, rootKPI, kpiEntityDtoToEntityMapping)
    }

    private fun storeKPI(
        repositoryEntity: RepositoryEntity,
        kpi: KPICreateDto,
        kpiEntityDtoToEntityMapping: MutableMap<KPICreateDto, KPIEntity>
    ) {
        // already visited
        if (kpiEntityDtoToEntityMapping.contains(kpi)) {
            return
        }

        if (kpi.hierarchyEdges.isEmpty()) {
            // leaf node
            val kpiEntity = kpi.toDbObject()
            kpiEntity.repository = repositoryEntity

            kpiEntityDtoToEntityMapping[kpi] = kpiEntity
            kpiRepository.save(kpiEntity)
            logger.info("Storing leaf node ${kpi.name}")
            return
        }

        for (childEdge in kpi.hierarchyEdges) {
            storeKPI(repositoryEntity, childEdge.to, kpiEntityDtoToEntityMapping)
        }

        val kpiEntity = kpi.toDbObject()
        kpiEntity.repository = repositoryEntity
        kpiEntityDtoToEntityMapping[kpi] = kpiEntity
        kpiRepository.save(kpiEntity)
        logger.info("Storing node ${kpi.name}")
        for (hierarchyEdge in kpi.hierarchyEdges) {
            val hierarchyEdgeEntity = KPIHierarchyEdgeEntity()
            hierarchyEdgeEntity.from = kpiEntity
            hierarchyEdgeEntity.to = kpiEntityDtoToEntityMapping[hierarchyEdge.to]
            hierarchyEdgeEntity.weight = hierarchyEdge.weight
            kpiHierarchyEdgeRepository.save(hierarchyEdgeEntity)
            logger.info("Storing edge ${hierarchyEdge.weight}: From ${hierarchyEdge.from.name} to ${hierarchyEdge.to.name}")
        }
    }

    private fun purgeAllKPIs(repositoryEntity: RepositoryEntity) {
        val kpiEntities: List<KPIEntity> = kpiRepository.findByRepository_Id(repositoryEntity.id!!)
        kpiRepository.deleteAll(kpiEntities)
    }

    fun findKPIById(id: UUID): KPIEntity? {
        return kpiRepository.findByIdOrNull(id)
    }

    fun findRootKPIByRepository(repositoryEntity: RepositoryEntity): KPIEntity? {
        val kpi = kpiRepository.findFirstByRepository_IdAndIsRootTrue(repositoryEntity.id!!)
        if (kpi.isEmpty) {
            return null
        }
        return kpi.get()
    }
}
