package de.fraunhofer.iem.dataprovider.taskManager

import de.fraunhofer.iem.dataprovider.configuration.DirectoryPathsProperties
import de.fraunhofer.iem.dataprovider.configuration.OpenCodeGitlabApiProperties
import de.fraunhofer.iem.dataprovider.kpi.service.KPIService
import de.fraunhofer.iem.dataprovider.logger.getLogger
import de.fraunhofer.iem.dataprovider.repository.service.RepositoryService
import de.fraunhofer.iem.dataprovider.taskManager.events.*
import de.fraunhofer.iem.dataprovider.taskManager.tasks.dataQuery.CloneGitTask
import de.fraunhofer.iem.dataprovider.taskManager.tasks.dataQuery.GetGitlabProjectTask
import de.fraunhofer.iem.dataprovider.taskManager.tasks.dataQuery.GetRepositoryDetailsTask
import de.fraunhofer.iem.dataprovider.taskManager.tasks.kpiCalculation.MetricsTask
import de.fraunhofer.iem.dataprovider.taskManager.tasks.tools.ort.OrtApiTask
import de.fraunhofer.iem.dataprovider.tool.service.ToolService
import de.fraunhofer.iem.dataprovider.toolRun.service.ToolRunService
import jakarta.annotation.PostConstruct
import jakarta.annotation.PreDestroy
import kotlinx.coroutines.CoroutineScope
import kotlinx.coroutines.Dispatchers
import kotlinx.coroutines.channels.Channel
import kotlinx.coroutines.launch
import org.springframework.stereotype.Component


/**
 * The Task Manager takes tasks and distributes them to
 * underlying workers. Internally it uses a channel
 * to manage incoming tasks.
 */
@Component
class TaskManager(
    private val directoryPathsProperties: DirectoryPathsProperties,
    private val openCodeGitlabConfiguration: OpenCodeGitlabApiProperties,
    private val toolRunService: ToolRunService,
    private val toolService: ToolService,
    private val repositoryService: RepositoryService,
    private val kpiService: KPIService,
) {

    // The used default dispatcher is ok for CPU-bound workloads. However,
    // if they block for a long time it's better to use a custom thread
    // pool solution.
    private val worker = Worker("Worker")

    // Should be used for long-running tasks, which DON'T use CPU power and
    // are non-blocking
    private val ioWorker = Worker("IO-Worker", CoroutineScope(Dispatchers.IO))
    // Should be used for long-running CPU heavy tasks, which potentially block.
    // private val customWorker = getCustomThreadpoolWorker()

    private val mainScope = CoroutineScope(Dispatchers.Default)
    private val channel: Channel<Event> = Channel()

    suspend fun addEvent(event: Event) {
        channel.send(event)
    }

    @PostConstruct
    private fun start() {
        mainScope.launch {
            for (task in channel) {
                onReceive(task)
            }
        }
    }

    @PreDestroy
    fun close() {
        channel.close()
        worker.close()
        ioWorker.close()
    }


    private val logger = getLogger(javaClass)
    private val groupTaskManager = GroupTaskManager(::onReceive)
    suspend fun onReceive(event: Event) {
        logger.info("[${Thread.currentThread().name}] add task called in Task Manager $event")

        when (event) {
            is RepoChangedEvent -> {
                ioWorker.addTask(
                    GetGitlabProjectTask(
                        event.repoId,
                        openCodeGitlabConfiguration,
                        ::addEvent,
                        repositoryService,
                    )
                )
            }

            is GetGitlabProjectDoneEvent -> {

                ioWorker.addTask(
                    CloneGitTask(
                        event.gitRepository,
                        event.repoId,
                        ::addEvent,
                        directoryPathsProperties.gitCloneTargetDirectory
                    )
                )
            }

            is GitCloneDoneEvent -> {

                val groupId = groupTaskManager.createTaskGroup(event.repoId)

//                    val occmdTask = OccmdTask(
//                        event.outputDirectory,
//                        config.toolResultsTargetDirectory,
//                        openCodeGitlabConfiguration,
//                        repositoryService,
//                        ::addEvent,
//                        event.repoId,
//                        groupId,
//                        toolRunService
//                    )
                ioWorker.addTask(
                    OrtApiTask(
                        ::addEvent,
                        event.repoId,
                        toolService,
                        repositoryService
                    )
                )

//                    groupTaskManager.addTaskToGroup(groupId, occmdTask.taskID)
//
//                    worker.addTask(occmdTask)
            }

            is GroupTaskDoneEvent -> {
                logger.info("GroupTaskDoneEvent received")
                groupTaskManager.taskInGroupFinished(event.groupId, event.taskId)
            }

            is GroupTasksDoneEvent -> {
                logger.info("Adding repository details task")
                worker.addTask(
                    GetRepositoryDetailsTask(
                        event.repoId,
                        openCodeGitlabConfiguration,
                        repositoryService,
                        ::addEvent
                    )
                )
            }

            is GetRepositoryDetailsDoneEvent -> {
                worker.addTask(
                    MetricsTask(
                        event.repoId,
                        repositoryService,
                        kpiService,
                        ::addEvent,
                        toolRunService
                    )
                )
            }

            is RecalculateAllKpisEvent -> {
                repositoryService.getAllRepositories().forEach { repo ->
                    worker.addTask(
                        MetricsTask(
                            repo.id!!,
                            repositoryService,
                            kpiService,
                            ::addEvent,
                            toolRunService
                        )
                    )
                }
            }

            is TaskFailedEvent -> {
                logger.info("task failed event received. ${event.throwable}")
            }

            is GroupTaskFailedEvent -> {
                logger.info("Group task failed event received. ${event.throwable} ")
                groupTaskManager.taskInGroupFinished(event.groupId, event.taskId)
            }

            else -> {
                logger.info("Received event without special handling associated $event")
            }
        }
        logger.info("[${Thread.currentThread().name}] add task finished in Task Manager $event")
    }

}
